import uuid
from enum import (
    Enum,
)

from sqlalchemy import (
    TIMESTAMP,
    Column,
    Integer,
    String,
)
from sqlalchemy.dialects.postgresql import (
    UUID,
)
from sqlalchemy.ext.declarative import (
    declarative_base,
)

Base = declarative_base()


class AuthType(Enum):
    CREDENTIAL = 1
    TOKEN = 2


class Authentication(Base):
    __tablename__ = "authentication"
    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4, unique=True, nullable=False)
    auth_type = Column(Integer)
    auth_uuid = Column(String)
    user_uuid = Column(String)
    user_id = Column(String)
    token = Column(String)
    created_at = Column(TIMESTAMP)
    updated_at = Column(TIMESTAMP)

    def __repr__(self):
        return (
            "<Authentication(uuid='{}', auth_type='{}',"
            "auth_uuid={}, created_at={}, updated_at={})>".format(  # pragma: no cover
                self.uuid, AuthType(self.auth_type), self.auth_uuid, self.created_at, self.updated_at
            )
        )



import json
import logging
import secrets
from datetime import (
    datetime,
)
from uuid import (
    uuid4,
)

from aiohttp import (
    ClientConnectorError,
    ClientResponse,
    ClientSession,
    web,
)
from sqlalchemy import (
    desc,
)
from sqlalchemy.orm import (
    sessionmaker,
)
from yarl import (
    URL,
)

from .database.models import (
    Authentication,
    AuthType,
)

logger = logging.getLogger(__name__)


async def register_credentials(request: web.Request) -> web.Response:
    """ Register User """
    try:
        content = await request.json()

        if "username" not in content or "password" not in content:
            return web.json_response({"error": "Wrong data. Provide username and password."}, status=400)
    except Exception:
        return web.json_response({"error": "Wrong data. Provide username and password."}, status=400)

    user_creation = await create_user_service_call(request)

    if user_creation.status == 200:
        resp_json = json.loads(user_creation.text)
        user_uuid = resp_json["uuid"]
        data = {"username": content["username"], "password": content["password"]}

        credentials_response = await create_credentials_call(request, data)

        if credentials_response.status == 200:
            resp_json_cred = json.loads(credentials_response.text)
            credential_uuid = resp_json_cred["credential_uuid"]
            token = secrets.token_hex(20)
            await create_authentication(
                request, token, content["username"], user_uuid, credential_uuid, AuthType.CREDENTIAL.value
            )
        else:
            return credentials_response

    return user_creation


async def register_token(request: web.Request) -> web.Response:
    """ Register User using token """
    try:
        content = await request.json()

        if "email" not in content:
            return web.json_response({"error": "Wrong data. Provide email."}, status=400)
    except Exception:
        return web.json_response({"error": "Wrong data. Provide email."}, status=400)

    user_creation = await create_user_service_call(request)

    if user_creation.status == 200:
        resp_json = json.loads(user_creation.text)
        user_uuid = resp_json["uuid"]

        token_response = await create_token_service_call(request)

        if token_response.status == 200:
            resp_json_cred = json.loads(token_response.text)
            await create_authentication(
                request,
                resp_json_cred["token"],
                content["email"],
                user_uuid,
                resp_json_cred["uuid"],
                AuthType.TOKEN.value,
            )
        return token_response

    return user_creation


async def credentials_login(request: web.Request) -> web.Response:
    """ Login User """
    resp, token = await validate_credentials(request)
    return resp


async def token_login(request: web.Request) -> web.Response:
    """ Login User """
    try:
        token = await _get_authorization_token(request)
    except Exception:
        return web.json_response({"error": "Please provide Token."}, status=400)

    session = sessionmaker(bind=request.app["db_engine"])

    s = session()

    r = s.query(Authentication).filter(Authentication.token == token).order_by(desc(Authentication.updated_at)).first()

    if r is not None:
        if r.auth_type == AuthType.TOKEN.value:
            token_response = await create_token_service_call(request)
            res = json.loads(token_response.text)
            r.token = res["token"]
            r.auth_uuid = res["uuid"]
            s.commit()
            s.close()
            return web.json_response({"token": res["token"]})

    s.close()

    return web.json_response({"error": "Please provide correct Token."}, status=400)


async def validate_credentials(request: web.Request):
    """ Login User """
    response = await validate_credentials_call(request)

    if response.status == 200:
        resp_json = json.loads(response.text)
        credential_uuid = resp_json["credential_uuid"]
        token = await get_credential_token(request, credential_uuid)
        return web.json_response({"token": token}), token

    return response, None


async def get_credential_token(request: web.Request, credential_uuid: str):
    """ Get User by Session token """
    session = sessionmaker(bind=request.app["db_engine"])

    s = session()

    r = (
        s.query(Authentication)
        .filter(Authentication.auth_uuid == credential_uuid)
        .order_by(desc(Authentication.updated_at))
        .first()
    )
    s.close()

    return r.token


async def get_user_from_token(request: web.Request, auth_type: AuthType = AuthType.TOKEN) -> web.Response:
    """ Get User by Session token """
    try:
        token = await _get_authorization_token(request)
    except Exception:
        return web.json_response({"error": "Please provide Token."}, status=400)

    return await get_token_user(request, token, auth_type)


async def get_token_user(request: web.Request, token: str, auth_type: AuthType):
    session = sessionmaker(bind=request.app["db_engine"])

    s = session()

    r = s.query(Authentication).filter(Authentication.token == token).order_by(desc(Authentication.updated_at)).first()
    s.close()

    if r is not None:
        if r.auth_type == auth_type.value:
            user_call_response = await get_user_call(request, r.user_uuid)
            return user_call_response

    return web.HTTPBadRequest(text="Please provide correct Token.")


async def get_user_from_credentials(request: web.Request) -> web.Response:
    resp, token = await validate_credentials(request)

    if resp.status == 200:
        return await get_token_user(request, token, AuthType.CREDENTIAL)
    return resp


async def validate_token(request: web.Request) -> web.Response:
    """ Get User by Session token """
    try:
        token = await _get_authorization_token(request)
    except Exception:
        return web.json_response({"error": "Please provide Token."}, status=400)

    session = sessionmaker(bind=request.app["db_engine"])

    s = session()

    r = s.query(Authentication).filter(Authentication.token == token).order_by(desc(Authentication.updated_at)).first()
    s.close()

    if r is not None:
        if r.auth_type == AuthType.TOKEN.value:
            token_resp = await validate_token_call(request)

            if token_resp.status == 200:
                return await get_user_call(request, r.user_uuid)

        if r.auth_type == AuthType.CREDENTIAL.value:
            return await get_user_call(request, r.user_uuid)

    return web.json_response({"error": "Please provide correct Token."}, status=400)


async def get_user_call(request: web.Request, user_uuid: str) -> web.Response:
    """ Get User by Session token """
    user_host = request.app["config"].user_service.host
    user_port = request.app["config"].user_service.port
    user_path = request.app["config"].user_service.path

    credential_url = URL(f"http://{user_host}:{user_port}{user_path}/{user_uuid}")

    headers = request.headers.copy()
    data = await request.read()

    return await service_call(method="GET", url=credential_url, data=data, headers=headers)


async def create_credentials_call(request: web.Request, data: dict, method: str = "POST") -> web.Response:
    """ Orchestrate discovery and microservice call """
    credential_host = request.app["config"].credential_service.host
    credential_port = request.app["config"].credential_service.port
    credential_path = request.app["config"].credential_service.path

    credential_url = URL(f"http://{credential_host}:{credential_port}{credential_path}")

    return await service_call(method=method, url=credential_url, data=json.dumps(data))


async def validate_credentials_call(request: web.Request):
    """ Validate Credentials Call """
    credential_host = request.app["config"].credential_service.host
    credential_port = request.app["config"].credential_service.port
    credential_path = request.app["config"].credential_service.path

    credential_url = URL(f"http://{credential_host}:{credential_port}{credential_path}/validate")

    headers = request.headers.copy()
    data = await request.read()

    return await service_call(method="POST", url=credential_url, data=data, headers=headers)


async def validate_token_call(request: web.Request):
    """ Validate Credentials Call """
    token_host = request.app["config"].token_service.host
    token_port = request.app["config"].token_service.port
    token_path = request.app["config"].token_service.path

    token_url = URL(f"http://{token_host}:{token_port}{token_path}/validate")

    headers = request.headers.copy()
    data = await request.read()

    return await service_call(method="POST", url=token_url, data=data, headers=headers)


async def create_token_service_call(request: web.Request):
    """ Register User """
    token_host = request.app["config"].token_service.host
    token_port = request.app["config"].token_service.port
    token_path = request.app["config"].token_service.path

    credential_url = URL(f"http://{token_host}:{token_port}{token_path}")

    headers = request.headers.copy()
    data = await request.read()

    return await service_call(method="POST", url=credential_url, data=data, headers=headers)


async def create_user_service_call(request: web.Request) -> web.Response:
    """ Register User """
    user_host = request.app["config"].user_service.host
    user_port = request.app["config"].user_service.port
    user_path = request.app["config"].user_service.path

    user_url = URL(f"http://{user_host}:{user_port}{user_path}")

    headers = request.headers.copy()
    data = await request.read()

    return await service_call(method="POST", url=user_url, data=data, headers=headers)


async def service_call(method: str, url: URL, data=None, headers=None):
    try:
        async with ClientSession() as session:
            async with session.request(headers=headers, method=method, url=url, data=data) as response:
                return await _clone_response(response)
    except ClientConnectorError:
        return web.json_response(
            {"error": "The requested endpoint is not available."}, status=web.HTTPServiceUnavailable.status_code
        )


# noinspection PyMethodMayBeStatic
async def _clone_response(response: ClientResponse) -> web.Response:
    return web.Response(
        body=await response.read(), status=response.status, reason=response.reason, headers=response.headers,
    )


async def create_authentication(
    request: web.Request, token: str, user_id: str, user_uuid: str, auth_uuid: str, auth_type: AuthType
):
    """ Orchestrate discovery and microservice call """
    session = sessionmaker(bind=request.app["db_engine"])

    s = session()

    now = datetime.now()
    uuid = uuid4()

    credential = Authentication(
        uuid=uuid,
        user_uuid=user_uuid,
        auth_uuid=auth_uuid,
        user_id=user_id,
        token=token,
        auth_type=auth_type,
        created_at=now,
        updated_at=now,
    )
    s.add(credential)
    s.commit()

    s.close()

    return uuid


async def _get_authorization_token(request: web.Request):
    try:
        headers = request.headers
        if "Authorization" in headers and "Bearer" in headers["Authorization"]:
            return headers["Authorization"].split()[1]
        else:
            raise Exception
    except Exception as e:
        raise e



class AuthConfigException(Exception):
    """Base Api Gateway Exception."""


class NoTokenException(AuthConfigException):
    """Exception to be raised when token is not available."""


class ApiGatewayConfigException(AuthConfigException):
    """Base config exception."""



import logging

from aiohttp import (
    web,
)
from aiomisc.service.aiohttp import (
    AIOHTTPService,
)
from sqlalchemy import (
    create_engine,
)

from .config import (
    AuthConfig,
)
from .database.models import (
    Base,
)
from .handler import (
    credentials_login,
    get_user_from_credentials,
    get_user_from_token,
    register_credentials,
    register_token,
    token_login,
    validate_token,
)

logger = logging.getLogger(__name__)


class AuthRestService(AIOHTTPService):
    def __init__(self, address: str, port: int, config: AuthConfig):
        self.config = config
        self.engine = None
        super().__init__(address, port)

    async def create_application(self) -> web.Application:
        app = web.Application()

        app["config"] = self.config
        self.engine = await self.create_engine()
        await self.create_database()

        app["db_engine"] = self.engine

        app.router.add_route("POST", "/auth/credentials", register_credentials)
        app.router.add_route("POST", "/auth/credentials/login", credentials_login)
        app.router.add_route("GET", "/auth/credentials", get_user_from_credentials)

        app.router.add_route("POST", "/auth/token", register_token)
        app.router.add_route("POST", "/auth/token/login", token_login)
        app.router.add_route("GET", "/auth/token", get_user_from_token)

        app.router.add_route("POST", "/auth/validate-token", validate_token)

        return app

    async def create_engine(self):
        DATABASE_URI = (
            f"postgresql+psycopg2://{self.config.database.user}:{self.config.database.password}@"
            f"{self.config.database.host}:{self.config.database.port}/{self.config.database.dbname}"
        )

        return create_engine(DATABASE_URI)

    async def create_database(self):
        Base.metadata.create_all(self.engine)



from __future__ import (
    annotations,
)

import abc
import collections
import os
import typing as t
from distutils import (
    util,
)
from pathlib import (
    Path,
)

import yaml

from .exceptions import (
    AuthConfigException,
)

REST = collections.namedtuple("Rest", "host port")
DATABASE = collections.namedtuple("Database", "dbname user password host port")
USER_SERVICE = collections.namedtuple("UserService", "host port path")
CREDENTIAL_SERVICE = collections.namedtuple("CredentialService", "host port path create_token")
TOKEN_SERVICE = collections.namedtuple("TokenService", "host port path create_token")

_ENVIRONMENT_MAPPER = {
    "rest.host": "AUTH_REST_HOST",
    "rest.port": "AUTH_REST_PORT",
    "database.dbname": "AUTH_DATABASE_NAME",
    "database.user": "AUTH_DATABASE_USER",
    "database.password": "AUTH_DATABASE_PASSWORD",
    "database.host": "AUTH_DATABASE_HOST",
    "database.port": "AUTH_DATABASE_PORT",
    "user-service.host": "AUTH_USER_SERVICE_REST_HOST",
    "user-service.port": "AUTH_USER_SERVICE_REST_PORT",
    "credential-service.host": "AUTH_CREDENTIAL_SERVICE_REST_HOST",
    "credential-service.port": "AUTH_CREDENTIAL_SERVICE_REST_PORT",
    "credential-service.path": "AUTH_CREDENTIAL_SERVICE_REST_PATH",
    "token-service.host": "AUTH_TOKEN_SERVICE_REST_HOST",
    "token-service.port": "AUTH_TOKEN_SERVICE_REST_PORT",
    "token-service.path": "AUTH_TOKEN_SERVICE_REST_PATH",
}

_PARAMETERIZED_MAPPER = {
    "rest.host": "auth_rest_host",
    "rest.port": "auth_rest_port",
    "database.database": "auth_database_name",
    "database.user": "auth_database_user",
    "database.password": "auth_database_password",
    "database.host": "auth_database_host",
    "database.port": "auth_database_port",
    "user-service.host": "auth_user_service_rest_host",
    "user-service.port": "auth_user_service_rest_port",
    "credential-service.host": "auth_credential_service_rest_host",
    "credential-service.port": "auth_credential_service_rest_port",
    "credential-service.path": "auth_credential_service_rest_path",
    "token-service.host": "auth_token_service_rest_host",
    "token-service.port": "auth_token_service_rest_port",
    "token-service.path": "auth_token_service_rest_path",
}


class AuthConfig(abc.ABC):
    """Api Gateway config class."""

    __slots__ = ("_services", "_path", "_data", "_with_environment", "_parameterized")

    def __init__(self, path: t.Union[Path, str], with_environment: bool = True, **kwargs):
        if isinstance(path, Path):
            path = str(path)
        self._services = {}
        self._path = path
        self._load(path)
        self._with_environment = with_environment
        self._parameterized = kwargs

    @staticmethod
    def _file_exit(path: str) -> bool:
        if os.path.isfile(path):
            return True
        return False

    def _load(self, path):
        if self._file_exit(path):
            with open(path) as f:
                self._data = yaml.load(f, Loader=yaml.FullLoader)
        else:
            raise AuthConfigException(f"Check if this path: {path} is correct")

    def _get(self, key: str, **kwargs: t.Any) -> t.Any:
        if key in _PARAMETERIZED_MAPPER and _PARAMETERIZED_MAPPER[key] in self._parameterized:
            return self._parameterized[_PARAMETERIZED_MAPPER[key]]

        if self._with_environment and key in _ENVIRONMENT_MAPPER and _ENVIRONMENT_MAPPER[key] in os.environ:
            if os.environ[_ENVIRONMENT_MAPPER[key]] in ["true", "True", "false", "False"]:
                return bool(util.strtobool(os.environ[_ENVIRONMENT_MAPPER[key]]))  # pragma: no cover
            return os.environ[_ENVIRONMENT_MAPPER[key]]

        def _fn(k: str, data: dict[str, t.Any]) -> t.Any:
            current, _, following = k.partition(".")

            part = data[current]
            if not following:
                return part

            return _fn(following, part)

        return _fn(key, self._data)

    @property
    def rest(self) -> REST:
        """Get the rest config.

        :return: A ``REST`` NamedTuple instance.
        """
        return REST(host=self._get("rest.host"), port=int(self._get("rest.port")))

    @property
    def database(self) -> DATABASE:
        """Get the rest config.

        :return: A ``REST`` NamedTuple instance.
        """
        return DATABASE(
            dbname=self._get("database.dbname"),
            user=self._get("database.user"),
            password=self._get("database.password"),
            host=self._get("database.host"),
            port=int(self._get("database.port")),
        )

    @property
    def user_service(self) -> USER_SERVICE:
        """Get the rest config.

        :return: A ``REST`` NamedTuple instance.
        """
        return USER_SERVICE(
            host=self._get("user-service.host"),
            port=int(self._get("user-service.port")),
            path=self._get("user-service.path"),
        )

    @property
    def credential_service(self) -> CREDENTIAL_SERVICE:
        """Get the rest config.

        :return: A ``REST`` NamedTuple instance.
        """
        return CREDENTIAL_SERVICE(
            host=self._get("credential-service.host"),
            port=int(self._get("credential-service.port")),
            path=str(self._get("credential-service.path")),
            create_token=self._get("credential-service.create-token"),
        )

    @property
    def token_service(self) -> TOKEN_SERVICE:
        """Get the rest config.

        :return: A ``REST`` NamedTuple instance.
        """
        return TOKEN_SERVICE(
            host=self._get("token-service.host"),
            port=int(self._get("token-service.port")),
            path=str(self._get("token-service.path")),
            create_token=self._get("token-service.create-token"),
        )



__author__ = """Clariteia Devs"""
__email__ = "devs@clariteia.com"
__version__ = "0.1.0"

from .config import (
    AuthConfig,
)
from .exceptions import (
    AuthConfigException,
)
from .launchers import (
    EntrypointLauncher,
)
from .service import (
    AuthRestService,
)
